(function()
{

    const {Cc, Ci}  = require("chrome");
    const log = require("ko/logging").getLogger("codeintel/feature/completions");
    const obsvc = Cc["@mozilla.org/observer-service;1"].getService(Ci.nsIObserverService);
    const codeintel = require("codeintel/codeintel");
    const service = require("codeintel/service");
    const koEditor = require("ko/editor");
    const prefs = require("ko/prefs");
    const Panel = require("./completions/panel");
    const w = require("ko/windows").getMain();
    const uuidGen = require('sdk/util/uuid');
    const _ = require("contrib/underscore");
    const partSvc   = Cc["@activestate.com/koPartService;1"].getService(Ci.koIPartService);
    const legacy = w.ko;
    const commands = require("ko/commands");
    const process = require("codeintel/process");
    const views = require("ko/views");
    const l = require("ko/locale").use();
    const platform = require("sdk/system").platform;
    const pathSep = platform == "winnt" ? ";" : ":";

    log.setLevel(10);

    var activePanel,
        activeView,
        activeLine;

    var languages;

    var uuid;
    var debounceDelay = prefs.getLong("codeintel.completions.debounce_delay");
    var maxResultsHard = prefs.getLong("codeintel.completions.max_results_hard");
    var invoking = false;

    var onPosChanged = (e) =>
    {
        if ( ! activePanel)
            return;

        var data = e.detail;
        if (data.char)
            return;

        if (activePanel.isOpen())
            activePanel.destroy();
    };

    /**
     * onCharModified fires whenever a single char was added or deleted
     * if a char was added data will hold the character, if a char was deleted
     * data will be empty
     */
    var onCharModified = (e) =>
    {
        if ( ! languages)
        {
            _updateLanguages();
            return;
        }

        var showWhileTyping = prefs.getBoolean("codeintel.completions.while_typing");

        if ( ! activePanel && ! showWhileTyping)
            return;

        // don't trigger when a completion is invoked
        if (invoking)
        {
            invoking = false;
            return;
        }

        // why another check for invoking in a slightly different way?
        // because race conditions
        if (activePanel && activePanel.invoking)
            return;

        var data = e.detail;
        var view = e.originalTarget;
        var lineNo = koEditor.getLineNumber();
        
        if ("_ciCompletions" in view && ! view._ciCompletions)
            return;
        
        if ( ! (view.language in languages))
        {
            view._ciCompletions = false;
            return;
        }
        
        var language = languages[view.language];
        var startPos = koEditor.getLineStartPos();
        var endPos = koEditor.getCursorPosition();
        var line = !! koEditor.getRange(startPos, endPos).trim();

        if (activePanel)
        {
            // Update current panel if our position didn't fall below the start pos
            // and we're still typing alphanumeric characters
            if (line && activePanel.isOpen() && activePanel.view == view && ((data && data.match(/[\w\$-]/)) || ! data))
            {
                var pos = view.scimoz.currentPos;
                if (pos >= activePanel.startPos)
                {
                    activePanel.update();
                    return;
                }
            }

            // update can cause the panel to get destroyed
            if (activePanel)
                activePanel.destroy();
        }

        if ( ! showWhileTyping)
            return;
        
        if ( ! line && ! language.completion_trigger_empty_lines)
            return;

        if (uuid && activeView == view && lineNo == activeLine)
            return; // a request is already in progress
        
        if (language.completion_trigger_blacklist)
        {
            var rx = new RegExp(`[${language.completion_trigger_blacklist}]`);
            if (data.match(rx))
                return;
        }

        activeView = view;
        activeLine = lineNo;

        uuid = uuidGen.uuid().toString();
        var requestUuid = uuid;

        this.checkForCompletions(view, requestUuid);
    };

    this.checkForCompletions = (view, requestUuid) => _checkForCompletions(view, requestUuid);

    var _updateLanguages = (callback) =>
    {
        service.getLanguages().then((_languages) =>
        {
            languages = _languages;
            if (callback)
                callback();
        });
    };

    var _checkForCompletions = _.throttle((view, requestUuid) =>
    {
        if ( ! view)
            view = require("ko/views").current().get();

        if ( ! languages)
        {
            _updateLanguages(_checkForCompletions.bind(this, view, requestUuid));
            return;
        }

        if ( ! (view.language in languages) ||
               languages[view.language].supports.indexOf(codeintel.FEATURE_COMPLETIONS) == -1)
        {
            view._ciCompletions = false;
            return;
        }

        view._ciCompletions = true;

        var curProject = partSvc.currentProject;
        var cwd = curProject ? curProject.liveDirectory : legacy.uriparse.URIToPath(legacy.places.getDirectory());
        var path = views.current().filePath;

        var importPaths = codeintel.getImportPaths(view.prefs, view.language, languages).join(pathSep);

        service.getCompletions(
        {
            buf: view.scimoz.text,
            pos: view.scimoz.currentPos,
            path: path,
            parentPath: cwd,
            importPaths: importPaths,
            language: view.language,
            limit: maxResultsHard
        })
        .then(onReceivedCompletions.bind(this, view, requestUuid))
        .catch(onReceivedCompletions.bind(this, view, requestUuid, false));
    }, debounceDelay);

    var onReceivedCompletions = (view, requestUuid, completions) =>
    {
        var _uuid = uuid;
        uuid = null;

        if (requestUuid != _uuid)
        {
            log.debug("Cancelling completion, uuid is no longer active");
            return;
        }
        
        if ( ! completions || (completions.entries.length === 0 && ! completions.signature))
        {
            if (activePanel && activePanel.isOpen())
            {
                activePanel.destroy();
                activePanel = null;
            }
            return;
        }
        
        if (activePanel &&
            activePanel.symbol == completions.symbol &&
            activePanel.view == view)
        {
            try
            {
                activePanel.update(completions);
            }
            catch (e)
            {
                log.exception(e, "Failed updating completion panel");
            }
        }
        else
        {
            try
            {
                activePanel = new Panel(view, languages[completions.language], completions);
                activePanel.onDestroy(() =>
                {
                    if (activePanel)
                        invoking = activePanel.invoking;
                    activePanel = null;
                });
            }
            catch (e)
            {
                log.exception(e, "Failed creating completion panel");
            }
        }
    };
    
    this.start = () =>
    {
        w.addEventListener("buffer_char_modified", onCharModified);
        w.addEventListener("buffer_pos_changed", onPosChanged);

        commands.register("triggerPrecedingCompletion", this.checkForCompletions,
        {
            isEnabled: () => true,
            label: l.get("completions.command.label")
        });
    };

    this.stop = () =>
    {
        w.removeEventListener("buffer_char_modified", onCharModified);
        w.removeEventListener("buffer_pos_changed", onPosChanged);
        commands.unregister("triggerPrecedingCompletion");
    };

}).apply(module.exports);
