(function() {

    const {Cc, Ci}  = require("chrome");
    const log = require("ko/logging").getLogger("codeintel/service/mediator");
    const codeintel = require("codeintel/codeintel");
    const process = require("codeintel/process");
    const jsonrpc = require("ko/jsonrpc");
    const legacy = Cc["@activestate.com/codeintel/legacy;1"].getService(Ci.koICodeintelLegacy);
    const timers = require("sdk/timers");
    const KoPromise = require("ko/promise");

    var cache = {};
    var age = {};

    const MAXAGE = 60 * 1000;

    var inProgress = {};
    var callbacks = {};

    //log.setLevel(10);

    /**
     * Generates an ID based on the name and args given
     * This only considers args that are numbers, booleans or strings
     *
     * @param   {string} name
     * @param   {array} args
     *
     * @returns {string} 
     */
    this.getId = (name, args) =>
    {
        var id = name;
        for (let arg of args)
        {
            let type = typeof arg;
            if (["string", "boolean", "number"].indexOf(type) == -1)
                continue;
            if (type == "string")
                id += arg.length;
            else
                id += arg;
        }
        return id;
    };

    /**
     * Register a mediated service
     * This takes care of all the boilerplate so the calling code only needs
     * to worry about the basics
     *
     * @param   {string} name           Endpoint
     * @param   {array} argIndex        Index of argument names (in the correct order)
     * @param   {object} optsDefault    Default options
     *
     * @returns {Function} Returns the mediated function
     */
    this.register = (name, argIndex, optsDefault) =>
    {
        return (args, opts = {}) =>
        {
            var _opts = Object.assign({}, optsDefault);
            _opts = Object.assign(_opts, opts);

            var _args = [];
            for (let k of argIndex)
            {
                let v = k in args ? args[k] : null;
                _args.push(v);
            }

            _opts.requestId = this.getId(name, _args);

            return this.mediate(name, _args, _opts);
        };
    };

    /**
     * Makes a request to the codeintel server
     *
     * @param   {Arguments} args [method, args]
     *
     * @returns {KoPromise}
     */
    this.request = (...args) =>
    {
        log.debug("request called");
        var client = codeintel.getClient();
        return client
            .request.apply(client.request, args)
            .catch((error) => { log.error(`Request failed with message: ${error.message} (${error.code})`, true /* no traceback */); });
    };

    /**
     * Makes a request to the codeintel server, if the codeintel server
     * fails to handle the request or has no results the request is instead
     * made to the legacy codeintel service
     *
     * @param   {Arguments} args [method, args]
     *
     * @returns {KoPromise}
     */
    this.fallbackRequest = (...args) =>
    {
        log.debug("fallbackRequest called");
        return new KoPromise((resolve, reject) =>
        {
            var fallback = () =>
            {
                this.legacyRequest.apply(this, args)
                .then(resolve)
                .catch(reject);
            };

            this.request.apply(this, args)
            .then((result) => {
                if ( ! result)
                    fallback();
                else
                    resolve(result);
            })
            .catch(fallback);
        });
    };

    /**
     * Makes a request to both the codeintel service as well as the legacy
     * codeintel service. The results of both requests are combined and then
     * returned via the KoPromise
     *
     * @param   {Arguments} args [method, args]
     *
     * @returns {KoPromise}
     */
    this.combinedRequest = (...args) =>
    {
        log.debug("combinedRequest called");
        return new KoPromise((resolve, reject) =>
        {
            var resultsReceived = 0;
            var result, resultLegacy;
            var combine = () =>
            {
                if (++resultsReceived != 2)
                    return;

                if (result && ! resultLegacy)
                {
                    resolve(result);
                    return;
                }

                if ( ! result && resultLegacy)
                {
                    resolve(resultLegacy);
                    return;
                }

                if ( ! result && ! resultLegacy)
                {
                    reject();
                    return;
                }

                if (Array.isArray(result))
                {
                    resolve(result.concat(resultLegacy));
                    return;
                }

                for (let k in resultLegacy)
                {
                    if ( ! resultLegacy.hasOwnProperty(k))
                        continue;

                    if (k in result)
                        continue;

                    result[k] = resultLegacy[k];
                }

                resolve(result);
            };

            this.request.apply(this, args)
            .then((data) => {
                result = data;
                combine();
            })
            .catch(combine);

            this.legacyRequest.apply(this, args)
            .then((data) => {
                resultLegacy = data;
                combine();
            })
            .catch(combine);
        });
    };

    /**
     * Makes a request to the legacy codeintel service
     * 
     * @param   {string} method 
     * @param   {object} args   
     * 
     * @returns {KoPromise} 
     */
    this.legacyRequest = (method, args = []) =>
    {
        log.debug("legacyRequest called");
        return new KoPromise((resolve, reject) =>
        {
            if ( ! (method in legacy))
            {
                log.error("method does not exist in koCodeintelLegacy: " + method);
                reject("Method does not exist");
                return;
            }

            var callback = (code, result) =>
            {
                if (code === 0)
                    resolve(JSON.parse(result));
                else
                    reject(result);
            };


            args.push(callback);

            try
            {
                legacy[method].apply(legacy, args);
            }
            catch (e)
            {
                log.exception(e, "Legacy request failed");
                reject();
            }
        });
    };

    /**
     * Mediate a request, this takes care of directing the request to the
     * proper request method, combining multiple requests of the same requestId,
     * caching, etc.
     * 
     * @param   {String} id   Service endpoint
     * @param   {Object} args 
     * @param   {Object} opts 
     * 
     * @returns {KoPromise} 
     */
    this.mediate = (id, args, opts = {}) =>
    {
        log.debug("Mediating request for " + id);

        // Flush old caches
        gc();

        var client = codeintel.getClient();

        opts.method = opts.method || false;
        opts.cache = opts.cache || false;
        opts.maxAge = opts.maxAge || false;

        if (args)
            args = [id, args];
        else
            args = [id];

        if ( ! opts.requestId)
        {
            log.warn(`Request does not have a requestId: ${id}`);
        }

        var requestId = opts.requestId || id;
        if ( ! (requestId in callbacks))
            callbacks[requestId] = [];

        // return cached entry if it exist and didnt expire
        if (opts.cache && requestId in cache && ( ! opts.maxAge || (Date.now() - age[requestId]) < opts.maxAge))
        {
            log.debug("Returning cached results");
            return new KoPromise((resolve, reject) =>
            {
                resolve.apply(null, cache[requestId]);
            });
        }

        // if a request id was supplied then see if we have any requests currently
        // in progress and if so add a callback handler rather than invoke the
        // request again
        if (requestId in inProgress && opts.cache)
        {
            log.debug("Request is already in progress, adding to callback chain");
            return new KoPromise((resolve, reject, each) =>
            {
                callbacks[requestId].push({resolve: resolve, reject: reject, each: each});
            });
        }

        if (process.get("status") != process.STATUS_STARTED)
        {
            if ( ! opts.waitingForProcess && ! opts.cache)
            {
                return new KoPromise((resolve, reject, each) =>
                {
                    log.warn("Cannot fulfill service request while codeintel process is down");
                    reject();
                });
            }

            timers.setTimeout(this.mediate.bind.apply(this.mediate, [null].concat(arguments)), 1000);

            if ( ! opts.waitingForProcess)
            {
                opts.waitingForProcess = true;
                log.warn("Waiting for codeintel process");
                return new KoPromise((resolve, reject, each) =>
                {
                    callbacks[requestId].push({resolve: resolve, reject: reject, each: each});
                });
            }

            return;
        }

        if (opts.cache)
            inProgress[requestId] = true;

        // Fire the actual request
        var result;
        if (opts.method == "fallback")
        {
            result = this.fallbackRequest.apply(this, args);
        }
        else if (opts.method == "legacy")
        {
            result = this.legacyRequest.apply(this, args);
        }
        else if (opts.method == "combined")
        {
            result = this.combinedRequest.apply(this, args);
        }
        else
        {
            result = this.request.apply(this, args);
        }

        // Handle callbacks for additional requests
        var callbackHandler = (type, args) =>
        {
            if (requestId in callbacks)
            {
                try
                {
                    for (let callback of callbacks[requestId])
                    {
                        log.debug("Calling callback for " + requestId);
                        callback[type].apply(null, args);
                    }
                }
                finally
                {
                    if (type != "each")
                    {
                        delete callbacks[requestId];
                    }
                }
            }

            if (opts.cache)
                delete inProgress[requestId];
        };
        
        result.each(() =>
        {
            callbackHandler("each", arguments);
        });

        result.then(() =>
        {
            if (opts.cache)
            {
                log.debug("Caching " + requestId);

                cache[requestId] = arguments;
                age[requestId] = Date.now();
            }

            log.debug("Resolving");
            callbackHandler("resolve", arguments);
        });

        result.catch(() =>
        {
            log.debug("Rejecting");
            callbackHandler("reject", arguments);
        });

        return result;
    };

    /**
     * Flush the cache
     *
     * @param   {String|Null} id    Only flushes cache for given id if defined
     *
     * @returns {Void}
     */
    this.flush = (id) =>
    {
        log.debug("flush " + id);
        if ( ! id)
        {
            age = {};
            cache = {};
        }
        else
        {
            if (id in cache)
                delete cache[id];
            if (id in age)
                delete age[id];
        }
    };

    /**
     * Garbage collector, flushes cache every MAXAGE ms
     * 
     * @returns {void} 
     */
    var gc = () =>
    {
        log.debug("gc");
        for (let id in age)
        {
            if (Date.now() - age[id] > (MAXAGE))
            {
                delete age[id];
                delete cache[id];
            }
        }
    };


}).apply(module.exports);